;; This file is part of Monocypher for CHICKEN
;; Copyright (c) 2020, Thomas Chust.  All rights reserved.
;;
;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions are
;; met:
;;
;; 1. Redistributions of source code must retain the above copyright
;;    notice, this list of conditions and the following disclaimer.
;;
;; 2. Redistributions in binary form must reproduce the above copyright
;;    notice, this list of conditions and the following disclaimer in the
;;    documentation and/or other materials provided with the
;;    distribution.
;;
;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;; A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
;; HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
;; LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
;; DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
;; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
;; (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
;; OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

(module (monocypher protobuf)
  (write-key read-key
   write-envelope read-envelope)
  (import
   scheme
   (chicken base)
   (only (chicken gc) set-finalizer!)
   (only (chicken blob) blob?)
   srfi-4
   srfi-12
   monocypher
   protobuf-encoding)

;; Auxiliary function to raise an I/O error.
(define (i/o-error location message . arguments)
  (abort
    (make-composite-condition
      (make-property-condition 'exn
        'message message
        'arguments arguments
        'location location
        'call-chain (get-call-chain))
      (make-property-condition 'i/o))))

;; Auxiliary function to skip over a protocol buffer field.
(define (skip location type port)
  (case type
    [(int*)  (read-int* port)]
    [(32bit) (read-fixed32 port)]
    [(64bit) (read-fixed64 port)]
    [(sized) (wipe! (read-sized-bytes port))]
    [else    (i/o-error location "invalid encoding")]))

;; Writes the given key in protocol buffer encoding to an output port.
(define (write-key k #!optional [port (current-output-port)])
  (write-tag/type 1 'sized port)
  (write-sized-string (key-algorithm k) port)
  (cond
    [(public-key-envelope k)
     => (lambda (pk)
          (write-tag/type 3 'sized port)
          (write-sized write-envelope pk port))]
    [(public-key k)
     => (lambda (pk)
          (write-tag/type 2 'sized port)
          (write-sized-bytes (blob->u8vector/shared pk) port))])
  (cond
    [(secret-key k)
     => (lambda (sk)
          (cond
            [(blob? sk)
             (write-tag/type 4 'sized port)
             (write-sized-bytes (blob->u8vector/shared sk) port)]
            [(envelope? sk)
             (write-tag/type 5 'sized port)
             (write-sized write-envelope sk port)]
            [else
             (error 'write-key "invalid secret key" sk)]))]))

;; Reads the protocol buffer encoding of a key from an input port.
(define (read-key #!optional [port (current-input-port)])
  (let loop ([algorithm #f] [pk #f] [sk #f])
    (let-values ([(tag type) (read-tag/type port)])
      (case tag
        [(#!eof)
         (if (and algorithm (or pk sk))
           (make-key algorithm pk sk)
           (i/o-error 'read-key "incomplete key"))]
        [(1)
         (if (eq? type 'sized)
           (loop (read-sized-string port) pk sk)
           (i/o-error 'read-key "invalid algorithm"))]
        [(2)
         (if (eq? type 'sized)
           (loop algorithm (u8vector->blob/shared (read-sized-bytes port)) sk)
           (i/o-error 'read-key "invalid public key"))]
        [(3)
         (if (eq? type 'sized)
           (loop algorithm (read-sized read-envelope port) sk)
           (i/o-error 'read-key "invalid public key"))]
        [(4)
         (if (eq? type 'sized)
           (loop algorithm pk (set-finalizer! (u8vector->blob/shared (read-sized-bytes port)) wipe!))
           (i/o-error 'read-key "invalid secret key"))]
        [(5)
         (if (eq? type 'sized)
           (loop algorithm pk (read-sized read-envelope port))
           (i/o-error 'read-key "invalid secret key"))]
        [else
         (skip 'read-key type port)
         (loop algorithm pk sk)]))))

;; Writes the given envelope in protocol buffer encoding to an output port.
(define (write-envelope e #!optional [port (current-output-port)])
  (write-tag/type 1 'sized port)
  (write-sized-string (envelope-algorithm e) port)
  (cond
    [(envelope-nonce e)
     => (lambda (n)
          (write-tag/type 2 'sized port)
          (write-sized-bytes n port))])
  (write-tag/type 3 'sized port)
  (write-sized-string (envelope-message e) port)
  (cond
    [(envelope-signature e)
     => (lambda (s)
          (write-tag/type 4 'sized port)
          (write-sized-bytes (blob->u8vector/shared s) port))]))

;; Reads the protocol buffer encoding of an envelope from an input port.
(define (read-envelope #!optional [port (current-input-port)])
  (let loop ([algorithm #f] [nonce #f] [message #f] [signature #f])
    (let-values ([(tag type) (read-tag/type port)])
      (case tag
        [(#!eof)
         (if (and algorithm message)
           (make-envelope algorithm nonce message signature)
           (i/o-error 'read-envelope "incomplete envelope"))]
        [(1)
         (if (eq? type 'sized)
           (loop (read-sized-string port) nonce message signature)
           (i/o-error 'read-envelope "invalid algorithm"))]
        [(2)
         (if (eq? type 'sized)
           (loop algorithm (read-sized-bytes port) message signature)
           (i/o-error 'read-envelope "invalid nonce"))]
        [(3)
         (if (eq? type 'sized)
           (loop algorithm nonce (read-sized-string port) signature)
           (i/o-error 'read-envelope "invalid message"))]
        [(4)
         (if (eq? type 'sized)
           (loop algorithm nonce message (u8vector->blob/shared (read-sized-bytes port)))
           (i/o-error 'read-envelope "invalid signature"))]
        [else
         (skip 'read-envelope type port)
         (loop algorithm nonce message signature)]))))

)
;; vim: set ft=scheme sts=2 sw=2 ts=4 et ai: ;;
