;; This file is part of Monocypher for CHICKEN
;; Copyright (c) 2020, Thomas Chust.  All rights reserved.
;;
;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions are
;; met:
;;
;; 1. Redistributions of source code must retain the above copyright
;;    notice, this list of conditions and the following disclaimer.
;;
;; 2. Redistributions in binary form must reproduce the above copyright
;;    notice, this list of conditions and the following disclaimer in the
;;    documentation and/or other materials provided with the
;;    distribution.
;;
;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;; A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
;; HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
;; LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
;; DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
;; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
;; (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
;; OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

(module monocypher
  (current-entropy-port stream-xor! stream-xor blob=?/O1 wipe!
   current-nonce-prefix generate-nonce nonce-seconds
   key? make-key key-algorithm public-key secret-key
   signed-key? boxed-key? make-public-key public-key-envelope
   check-key generate-key derive-key
   envelope envelope? make-envelope envelope-algorithm envelope-nonce envelope-message envelope-signature
   valid-envelope?
   hash-algorithm
   hash
   random-stream-algorithm random-stream-keybytes random-stream-noncebytes
   generate-random-stream-key derive-random-stream-key open-random-stream
   symmetric-sign-algorithm symmetric-sign-keybytes
   generate-symmetric-sign-key derive-symmetric-sign-key symmetric-sign symmetric-verify symmetric-key-sign symmetric-key-verify
   symmetric-box-algorithm symmetric-box-keybytes symmetric-box-noncebytes
   generate-symmetric-box-key derive-symmetric-box-key symmetric-box symmetric-unbox symmetric-key-box symmetric-key-unbox
   asymmetric-sign-algorithm asymmetric-sign-publickeybytes asymmetric-sign-secretkeybytes
   generate-asymmetric-sign-key derive-asymmetric-sign-key asymmetric-sign asymmetric-verify asymmetric-key-sign asymmetric-key-verify
   asymmetric-box-algorithm asymmetric-box-hiddenkeybytes asymmetric-box-publickeybytes asymmetric-box-secretkeybytes asymmetric-box-noncebytes
   generate-asymmetric-box-key derive-asymmetric-box-key hide-asymmetric-box-key reveal-asymmetric-box-key exchange-random-stream-key exchange-symmetric-box-key asymmetric-box asymmetric-unbox asymmetric-key-box asymmetric-key-unbox)
  (import
   scheme
   (chicken base)
   (chicken foreign)
   (chicken blob)
   (chicken port)
   (only (chicken random) random-bytes)
   (only (chicken time) current-seconds current-process-milliseconds)
   (only (chicken fixnum) fx+ fx- fx* fx= fx< fx> fx>=)
   (only (chicken memory) move-memory!)
   (only (chicken io) read-string read-byte)
   (only (chicken gc) set-finalizer!)
   srfi-12
   srfi-4)

(foreign-declare "#include \"monocypher.h\"")

(define current-entropy-port
  (make-parameter
   (make-input-port
    ;read-char
    (lambda ()
      (string-ref (random-bytes (make-string 1)) 0))
    ;char-ready?
    (lambda ()
      #t)
    ;close
    void
    ;peek-char
    #f
    ;read-string!
    (lambda (port len buf ofs)
      (if (zero? ofs)
          (begin
            (random-bytes buf len)
            len)
          (let ([tmp (make-string len)])
            (random-bytes tmp len)
            (move-memory! tmp buf len 0 ofs)
            len))))))

(define (i/o-error location message . arguments)
  (abort
    (make-composite-condition
      (make-property-condition 'exn
        'message message
        'arguments arguments
        'location location
        'call-chain (get-call-chain))
      (make-property-condition 'i/o))))

(define (stream-xor! buffer #!optional [port (current-input-port)])
  (let* ([len (string-length buffer)]
         [key (read-string len port)])
    (when (or (eof-object? key)
              (< (string-length key) len))
      (i/o-error 'stream-xor! "key source depleted" port))
    ((foreign-lambda*
      void ([nonnull-scheme-pointer c_] [nonnull-scheme-pointer k_] [unsigned-integer n])
      "unsigned char *c = (unsigned char *)c_;\n"
      "const unsigned char *k = (unsigned char *)k_;\n"
      "while (n-- > 0) *c++ ^= *k++;\n")
     buffer key len))
  buffer)

(define (stream-xor plain #!optional [port (current-input-port)])
  (stream-xor! (string-copy plain) port))

(define (blob=?/O1 a b)
  (let ([na (blob-size a)]
        [nb (blob-size b)])
    (cond
      [(= na nb 0)
       #t]
      [(= na nb 16)
       (not ((foreign-lambda bool "crypto_verify16" nonnull-blob nonnull-blob) a b))]
      [(= na nb 32)
       (not ((foreign-lambda bool "crypto_verify32" nonnull-blob nonnull-blob) a b))]
      [(= na nb 64)
       (not ((foreign-lambda bool "crypto_verify64" nonnull-blob nonnull-blob) a b))]
      [else
       (type-error 'blob=?/O1 "expected two blobs of length 0, 16, 32, or 64" a b)])))

(define (wipe! v)
  (cond
    [(key? v)
     (cond [(secret-key v) => wipe!])]
    [(envelope? v)
     (cond [(envelope-message v) => wipe!])]
    [(u8vector? v)
     (wipe! (u8vector->blob/shared v))]
    [(s8vector? v)
     (wipe! (s8vector->blob/shared v))]
    [(u16vector? v)
     (wipe! (u16vector->blob/shared v))]
    [(s16vector? v)
     (wipe! (s16vector->blob/shared v))]
    [(u32vector? v)
     (wipe! (u32vector->blob/shared v))]
    [(s32vector? v)
     (wipe! (s32vector->blob/shared v))]
    [(u64vector? v)
     (wipe! (u64vector->blob/shared v))]
    [(s64vector? v)
     (wipe! (s64vector->blob/shared v))]
    [(f32vector? v)
     (wipe! (f32vector->blob/shared v))]
    [(f64vector? v)
     (wipe! (f64vector->blob/shared v))]
    [else
     ((foreign-lambda void "crypto_wipe" nonnull-scheme-pointer size_t)
      v
      (cond
        [(string? v)
         (string-length v)]
        [(blob? v)
         (blob-size v)]
        [else
         (type-error 'wipe! "expected a key, envelope, string, blob, or number vector" v)]))]))

(define current-nonce-prefix
  (make-parameter #f))

(define (generate-nonce location nonce-bytes #!optional [port (current-entropy-port)])
  (let ([prefixbytes (fx- nonce-bytes 8)]
        [n (make-u8vector nonce-bytes 0)])
    (cond
      [(fx< prefixbytes 0)
       (error location "invalid nonce length" nonce-bytes)]
      [(fx> prefixbytes 0)
       (let ([p (current-nonce-prefix)])
         (unless p
           (set! p (make-u8vector prefixbytes))
           (unless (fx= (read-u8vector! #f p port) prefixbytes)
             (i/o-error location "entropy source depleted" port))
           (current-nonce-prefix p))
         (move-memory! p n prefixbytes))])
    (let next ([i (fx- (u8vector-length n) 1)]
               [t (+ (* (current-seconds) 1000) (modulo (current-process-milliseconds) 1000))])
      (unless (zero? t)
        (u8vector-set! n i (modulo t 256))
        (next (fx- i 1) (quotient t 256))))
    n))

(define (nonce-seconds n)
  (let next ([i (fx- (u8vector-length n) 7)]
             [t 0])
    (if (fx< i (u8vector-length n))
      (next (fx+ i 1) (+ (* t 256) (u8vector-ref n i)))
      (/ t 1000))))

(define-record-type key
  (make-key algorithm public secret) key?
  [algorithm key-algorithm]
  [public public-key*]
  [secret secret-key])

(define (public-key k)
  (and-let* ([pk (public-key* k)])
    (if (envelope? pk)
      (string->blob (envelope-message pk))
      pk)))

(define (public-key-envelope k)
  (and-let* ([pk (public-key* k)]
             [(envelope? pk)])
    pk))

(define (type-error location message . arguments)
  (abort
    (make-composite-condition
      (make-property-condition 'exn
        'message message
        'arguments arguments
        'location location
        'call-chain (get-call-chain))
      (make-property-condition 'type))))

(define (check-key location algorithm v)
  (if (and (key? v) (equal? (key-algorithm v) algorithm))
    v
    (type-error location (string-append "expected " algorithm " key") v)))

(define (signed-key? v)
  (and (key? v) (envelope? (public-key* v))))

(define (boxed-key? v)
  (and (key? v) (envelope? (secret-key v))))

(define (make-public-key k)
  (make-key (key-algorithm k) (public-key* k) #f))

(define ((generate-key location algorithm secret-keybytes secret->public) #!optional [port (current-entropy-port)])
  (let ([k (read-u8vector secret-keybytes port)])
    (if (or (eof-object? k) (fx< (u8vector-length k) secret-keybytes))
      (i/o-error location "entropy source depleted" port)
      (set! k (u8vector->blob/shared k)))
    (set-finalizer! k wipe!)
    (make-key algorithm (secret->public k) k)))

(define ((derive-key* location secret-keybytes wrap-secret) password salt #!key [blocks 262144] [iterations 8])
  (unless (fx>= (u8vector-length salt) 8)
    (error location "invalid salt" salt))
  (unless (fx>= blocks 8)
    (error location "invalid number of blocks" blocks))
  (unless (fx>= iterations 1)
    (error location "invalid number of iterations" iterations))
  (let ([k (make-blob secret-keybytes)]
        [work-area (make-blob (fx* 1024 blocks))])
    ((foreign-lambda*
       void ([nonnull-blob hash] [unsigned-integer32 hash_size]
	     [nonnull-blob work_area] [unsigned-integer32 blocks] [unsigned-integer32 iterations]
	     [nonnull-scheme-pointer password] [unsigned-integer32 password_size]
	     [nonnull-u8vector salt] [unsigned-integer32 salt_size])
       "crypto_argon2_config config = { CRYPTO_ARGON2_I, blocks, iterations, 1 };"
       "crypto_argon2_inputs inputs = { password, salt, password_size, salt_size };"
       "crypto_argon2_extras extras = { NULL, NULL, 0, 0 };"
       "crypto_argon2(hash, hash_size, work_area, config, inputs, extras);")
      k secret-keybytes
      work-area blocks iterations
      password (string-length password)
      salt (u8vector-length salt))
    (wrap-secret k)))

(define (derive-key location algorithm secret-keybytes secret->public)
  (derive-key*
   location secret-keybytes
   (lambda (k)
     (set-finalizer! k wipe!)
     (make-key algorithm (secret->public k) k))))

(define-record-type envelope
  (make-envelope algorithm nonce message signature) envelope?
  [algorithm envelope-algorithm]
  [nonce envelope-nonce]
  [message envelope-message]
  [signature envelope-signature])

(define (valid-envelope? algorithm nonce-bytes signature-bytes v)
  (and (envelope? v)
       (equal? (envelope-algorithm v) algorithm)
       (cond
         [(number? nonce-bytes)
          (fx= (u8vector-length (envelope-nonce v)) nonce-bytes)]
         [nonce-bytes
          #t]
         [else
          (not (envelope-nonce v))])
       (fx= (blob-size (envelope-signature v)) signature-bytes)))

(define hash-algorithm "Blake2b")

(define hash-bytes 64)

(define (hash m)
  (let ([h (make-blob hash-bytes)])
    ((foreign-lambda
       void "crypto_blake2b"
       nonnull-blob size_t nonnull-scheme-pointer size_t)
     h (blob-size h) m (string-length m))
    h))

(define random-stream-algorithm "XChaCha20")
(define random-stream-keybytes 32)
(define random-stream-noncebytes 24)

(define random-stream-blocksize 16)

(define-values (generate-random-stream-key derive-random-stream-key)
  (letrec ([secret->public (constantly #f)])
    (values
      (generate-key
        'generate-random-stream-key random-stream-algorithm random-stream-keybytes
        secret->public)
      (derive-key
        'derive-random-stream-key random-stream-algorithm random-stream-keybytes
        secret->public))))

(define (open-random-stream k #!key [nonce (generate-nonce 'open-random-stream random-stream-noncebytes)] [limit (expt 2 30)])
  (set! k
    (secret-key (check-key 'open-random-stream random-stream-algorithm k)))
  (unless (fx= (blob-size k) random-stream-keybytes)
    (error 'open-random-stream "invalid key" k))
  (unless (fx= (u8vector-length nonce) random-stream-noncebytes)
    (error 'open-random-stream "invalid nonce" nonce))
  (unless limit
    (set! limit +inf.0))
  (let ([ctr 0]
        [buf (make-string random-stream-blocksize)]
        [pos random-stream-blocksize])
    (make-input-port
     ;read-char
     (lambda ()
       (when (and (fx>= pos random-stream-blocksize) (positive? limit))
         (wipe! buf)
         (set! ctr
           ((foreign-lambda
             unsigned-integer64 "crypto_chacha20_x"
             nonnull-scheme-pointer nonnull-scheme-pointer size_t
             nonnull-blob nonnull-u8vector unsigned-integer64)
            buf buf random-stream-blocksize k nonce ctr))
         (set! limit (- limit random-stream-blocksize))
         (set! pos 0))
       (if (fx< pos random-stream-blocksize)
         (let ([c (string-ref buf pos)])
           (set! pos (fx+ pos 1))
           c)
         #!eof))
     ;char-ready?
     (lambda ()
       (or (fx< pos random-stream-blocksize)
           (positive? limit)))
     ;close
     (lambda ()
       (wipe! buf)))))

(define symmetric-sign-algorithm "Poly1305")
(define symmetric-sign-keybytes 32)

(define symmetric-sign-bytes 16)

(define crypto-poly1305
  (foreign-lambda
    void "crypto_poly1305"
    nonnull-blob nonnull-scheme-pointer size_t nonnull-blob))

(define-values (generate-symmetric-sign-key derive-symmetric-sign-key)
  (letrec ([secret->public (constantly #f)])
    (values
      (generate-key
        'generate-symmetric-sign-key symmetric-sign-algorithm symmetric-sign-keybytes
        secret->public)
      (derive-key
        'derive-symmetric-sign-key symmetric-sign-algorithm symmetric-sign-keybytes
        secret->public))))

(define (symmetric-sign k)
  (set! k
    (secret-key (check-key 'symmetric-sign symmetric-sign-algorithm k)))
  (unless (fx= (blob-size k) symmetric-sign-keybytes)
    (error 'symmetric-sign "invalid key" k))
  (lambda (m #!key aad [nonce (generate-nonce 'symmetric-sign 8)])
    (let* ([m-bytes (string-length m)]
           [aad-bytes (if aad (string-length aad) 0)]
           [nonce-bytes (if nonce (u8vector-length nonce) 0)]
           [ofs 0]
           [buf (if (or aad nonce) (make-string (+ m-bytes aad-bytes nonce-bytes)) m)]
           [mac (make-blob symmetric-sign-bytes)])
      (when nonce
        (move-memory! nonce buf nonce-bytes 0 ofs)
        (set! ofs (fx+ ofs nonce-bytes)))
      (when aad
        (move-memory! aad buf aad-bytes 0 ofs)
        (set! ofs (fx+ ofs aad-bytes)))
      (when (or aad nonce)
        (move-memory! m buf m-bytes 0 ofs))
      (crypto-poly1305 mac buf (string-length buf) k)
      (make-envelope symmetric-sign-algorithm nonce m mac))))

(define (symmetric-verify k)
  (set! k
    (secret-key (check-key 'symmetric-verify symmetric-sign-algorithm k)))
  (unless (fx= (blob-size k) symmetric-sign-keybytes)
    (error 'symmetric-verify "invalid key" k))
  (lambda (e #!key aad)
    (and
      (valid-envelope? symmetric-sign-algorithm #t symmetric-sign-bytes e)
      (let* ([m (envelope-message e)]
             [m-bytes (string-length m)]
             [aad-bytes (if aad (string-length aad) 0)]
             [nonce (envelope-nonce e)]
             [nonce-bytes (if nonce (u8vector-length nonce) 0)]
             [ofs 0]
             [buf (if (or aad nonce) (make-string (+ m-bytes aad-bytes nonce-bytes)) m)]
             [mac (make-blob symmetric-sign-bytes)])
        (when nonce
          (move-memory! nonce buf nonce-bytes 0 ofs)
          (set! ofs (fx+ ofs nonce-bytes)))
        (when aad
          (move-memory! aad buf aad-bytes 0 ofs)
          (set! ofs (fx+ ofs aad-bytes)))
        (when (or aad nonce)
          (move-memory! m buf m-bytes 0 ofs))
        (crypto-poly1305 mac buf (string-length buf) k)
        (and
          (blob=?/O1 mac (envelope-signature e))
          m)))))

(define (symmetric-key-sign k)
  (let ([sign (symmetric-sign k)])
    (lambda (k . signargs)
      (let ([algorithm (key-algorithm k)])
        (make-key
          algorithm
          (apply sign (blob->string (public-key k)) #:aad algorithm signargs)
          (secret-key k))))))

(define (symmetric-key-verify k)
  (let ([verify (symmetric-verify k)])
    (lambda (k . verifyargs)
      (and-let* ([algorithm (key-algorithm k)]
                 [pk (apply verify (public-key* k) #:aad algorithm verifyargs)])
        (make-key
          algorithm
          (string->blob pk)
          (secret-key k))))))

(define symmetric-box-algorithm (string-append random-stream-algorithm "/" symmetric-sign-algorithm))
(define symmetric-box-keybytes 32)
(define symmetric-box-noncebytes 24)

(define-values (generate-symmetric-box-key derive-symmetric-box-key)
  (letrec ([secret->public (constantly #f)])
    (values
      (generate-key
        'generate-symmetric-box-key symmetric-box-algorithm symmetric-box-keybytes
        secret->public)
      (derive-key
        'derive-symmetric-box-key symmetric-box-algorithm symmetric-box-keybytes
        secret->public))))

(define (symmetric-box k)
  (set! k
    (secret-key (check-key 'symmetric-box symmetric-box-algorithm k)))
  (unless (fx= (blob-size k) symmetric-box-keybytes)
    (error 'symmetric-box "invalid key" k))
  (lambda (m #!key aad [nonce (generate-nonce 'symmetric-box symmetric-box-noncebytes)])
    (unless (fx= (u8vector-length nonce) symmetric-box-noncebytes)
      (error 'symmetric-box "invalid nonce" nonce))
    (let ([c (make-string (string-length m))]
          [mac (make-blob symmetric-sign-bytes)])
      ((foreign-lambda
         void "crypto_aead_lock"
         nonnull-scheme-pointer nonnull-blob
         nonnull-blob nonnull-u8vector
         scheme-pointer size_t
         nonnull-scheme-pointer size_t)
       c mac k nonce
       aad (if aad (string-length aad) 0) m (string-length m))
      (make-envelope symmetric-box-algorithm nonce c mac))))

(define (symmetric-unbox k)
  (set! k
    (secret-key (check-key 'symmetric-unbox symmetric-box-algorithm k)))
  (unless (fx= (blob-size k) symmetric-box-keybytes)
    (error 'symmetric-unbox "invalid key" k))
  (lambda (e #!key aad)
    (and
      (valid-envelope? symmetric-box-algorithm symmetric-box-noncebytes symmetric-sign-bytes e)
      (let* ([c (envelope-message e)]
             [m (make-string (string-length c))])
        (and
          (not
           ((foreign-lambda
              bool "crypto_aead_unlock"
              nonnull-scheme-pointer nonnull-blob
              nonnull-blob nonnull-u8vector
              scheme-pointer size_t
              nonnull-scheme-pointer size_t)
            m (envelope-signature e) k (envelope-nonce e)
	    aad (if aad (string-length aad) 0) c (string-length c)))
          (set-finalizer! m wipe!))))))

(define (symmetric-key-box k)
  (letrec ([box (symmetric-box k)])
    (lambda (k . boxargs)
      (let ([algorithm (key-algorithm k)]
            [sk (secret-key k)])
        (if sk
          (make-key
            algorithm
            (public-key* k)
            (apply box (set-finalizer! (blob->string sk) wipe!) #:aad algorithm boxargs))
          k)))))

(define (symmetric-key-unbox k)
  (letrec ([unbox (symmetric-unbox k)])
    (lambda (k . boxargs)
      (let ([algorithm (key-algorithm k)]
            [sk (secret-key k)])
        (if (envelope? sk)
          (and-let* ([sk (apply unbox sk #:aad algorithm boxargs)])
            (make-key
              algorithm
              (public-key* k)
              (set-finalizer! (string->blob sk) wipe!)))
          k)))))

(define asymmetric-sign-keyalgorithm "EdDSA/Curve25519")
(define asymmetric-sign-algorithm (string-append asymmetric-sign-keyalgorithm "/" hash-algorithm))
(define asymmetric-sign-seedbytes 32)
(define asymmetric-sign-publickeybytes 32)
(define asymmetric-sign-secretkeybytes 64)

(define asymmetric-sign-bytes 64)

(define (make-asymmetric-sign-key seed)
  (let ([pk (make-blob asymmetric-sign-publickeybytes)]
	[sk (make-blob asymmetric-sign-secretkeybytes)])
    ((foreign-lambda
      void "crypto_eddsa_key_pair"
      nonnull-scheme-pointer nonnull-scheme-pointer nonnull-scheme-pointer)
     sk pk seed)
    (make-key asymmetric-sign-keyalgorithm pk (set-finalizer! sk wipe!))))

(define (generate-asymmetric-sign-key #!optional [port (current-entropy-port)])
  (let ([s (read-u8vector asymmetric-sign-seedbytes port)])
    (if (or (eof-object? s) (fx< (u8vector-length s) asymmetric-sign-seedbytes))
      (i/o-error 'generate-asymmetric-sign-key "entropy source depleted" port)
      (set! s (u8vector->blob/shared s)))
    (make-asymmetric-sign-key s)))

(define derive-asymmetric-sign-key
  (derive-key* 'derive-asymmetric-sign-key asymmetric-sign-seedbytes make-asymmetric-sign-key))

(define (asymmetric-sign sk)
  (define pk #f)
  (check-key 'asymmetric-sign asymmetric-sign-keyalgorithm sk)
  (set! pk (public-key sk))
  (set! sk (secret-key sk))
  (unless (or (not pk) (fx= (blob-size pk) asymmetric-sign-publickeybytes))
    (error 'asymmetric-sign "invalid public key" pk))
  (unless (fx= (blob-size sk) asymmetric-sign-secretkeybytes)
    (error 'asymmetric-sign "invalid secret key" sk))
  (lambda (m #!key aad [nonce (generate-nonce 'asymmetric-sign 8)])
    (let* ([m-bytes (string-length m)]
           [aad-bytes (if aad (string-length aad) 0)]
           [nonce-bytes (if nonce (u8vector-length nonce) 0)]
           [ofs 0]
           [buf (if (or aad nonce) (make-string (+ m-bytes aad-bytes nonce-bytes)) m)]
           [sig (make-blob asymmetric-sign-bytes)])
      (when nonce
        (move-memory! nonce buf nonce-bytes 0 ofs)
        (set! ofs (fx+ ofs nonce-bytes)))
      (when aad
        (move-memory! aad buf aad-bytes 0 ofs)
        (set! ofs (fx+ ofs aad-bytes)))
      (when (or aad nonce)
        (move-memory! m buf m-bytes 0 ofs))
      ((foreign-lambda
         void "crypto_eddsa_sign"
         nonnull-blob nonnull-blob nonnull-scheme-pointer size_t)
       sig sk buf (string-length buf))
      (make-envelope asymmetric-sign-algorithm nonce m sig))))

(define (asymmetric-verify pk)
  (set! pk
    (public-key (check-key 'asymmetric-verify asymmetric-sign-keyalgorithm pk)))
  (unless (fx= (blob-size pk) asymmetric-sign-publickeybytes)
    (error 'asymmetric-verify "invalid public key" pk))
  (lambda (e #!key aad)
    (and
      (valid-envelope? asymmetric-sign-algorithm #t asymmetric-sign-bytes e)
      (let* ([m (envelope-message e)]
             [m-bytes (string-length m)]
             [aad-bytes (if aad (string-length aad) 0)]
             [nonce (envelope-nonce e)]
             [nonce-bytes (if nonce (u8vector-length nonce) 0)]
             [ofs 0]
             [buf (if (or aad nonce) (make-string (+ m-bytes aad-bytes nonce-bytes)) m)])
        (when nonce
          (move-memory! nonce buf nonce-bytes 0 ofs)
          (set! ofs (fx+ ofs nonce-bytes)))
        (when aad
          (move-memory! aad buf aad-bytes 0 ofs)
          (set! ofs (fx+ ofs aad-bytes)))
        (when (or aad nonce)
          (move-memory! m buf m-bytes 0 ofs))
        (and
          (not
            ((foreign-lambda
               bool "crypto_eddsa_check"
               nonnull-blob nonnull-blob nonnull-scheme-pointer size_t)
             (envelope-signature e) pk buf (string-length buf)))
          m)))))

(define (asymmetric-key-sign sk)
  (let ([sign (asymmetric-sign sk)])
    (lambda (k . signargs)
      (let ([algorithm (key-algorithm k)])
        (make-key
          algorithm
          (apply sign (blob->string (public-key k)) #:aad algorithm signargs)
          (secret-key k))))))

(define (asymmetric-key-verify pk)
  (let ([verify (asymmetric-verify pk)])
    (lambda (k . verifyargs)
      (and-let* ([algorithm (key-algorithm k)]
                 [pk (apply verify (public-key* k) #:aad algorithm verifyargs)])
        (make-key
          algorithm
          (string->blob pk)
          (secret-key k))))))

(define asymmetric-box-keyalgorithm "X25519/HChaCha20")
(define asymmetric-box-algorithm (string-append asymmetric-box-keyalgorithm "/" symmetric-box-algorithm))
(define asymmetric-box-hiddenkeybytes 32)
(define asymmetric-box-publickeybytes 32)
(define asymmetric-box-secretkeybytes 32)
(define asymmetric-box-noncebytes symmetric-box-noncebytes)

(define crypto-key-exchange
  (foreign-lambda
    void "crypto_x25519"
    nonnull-blob nonnull-blob nonnull-blob))

(define-values (generate-asymmetric-box-key derive-asymmetric-box-key)
  (letrec ([secret->public
            (lambda (sk)
              (let ([pk (make-blob asymmetric-box-publickeybytes)])
                ((foreign-lambda
                   void "crypto_x25519_public_key"
                   nonnull-blob nonnull-blob)
                 pk sk)
                pk))])
    (values
      (generate-key
        'generate-asymmetric-box-key asymmetric-box-keyalgorithm asymmetric-box-secretkeybytes
        secret->public)
      (derive-key
        'derive-asymmetric-box-key asymmetric-box-keyalgorithm asymmetric-box-secretkeybytes
        secret->public))))

(define (hide-asymmetric-box-key sk #!key [tweak (read-byte (current-entropy-port))])
  (set! sk
    (secret-key (check-key 'exchange-random-stream-key asymmetric-box-keyalgorithm sk)))
  (let ([curve (make-blob asymmetric-box-publickeybytes)]
	[hk (make-blob asymmetric-box-hiddenkeybytes)])
    ((foreign-lambda void "crypto_x25519_dirty_fast" nonnull-blob nonnull-blob)
     curve sk)
    (and
     (zero?
      ((foreign-lambda int "crypto_elligator_rev" nonnull-blob nonnull-blob unsigned-byte)
       hk curve tweak))
     hk)))

(define (reveal-asymmetric-box-key hk)
  (unless (fx= (blob-size hk) asymmetric-box-hiddenkeybytes)
    (error 'reveal-asymmetric-box-key "invalid hidden key" hk))
  (let ([curve (make-blob asymmetric-box-publickeybytes)])
    ((foreign-lambda void "crypto_elligator_map" nonnull-blob nonnull-blob)
     curve hk)
    (make-key asymmetric-box-keyalgorithm curve #f)))

(define (exchange-random-stream-key pk sk)
  (set! pk
    (public-key (check-key 'exchange-random-stream-key asymmetric-box-keyalgorithm pk)))
  (set! sk
    (secret-key (check-key 'exchange-random-stream-key asymmetric-box-keyalgorithm sk)))
  (unless (fx= (blob-size pk) asymmetric-box-publickeybytes)
    (error 'exchange-random-stream-key "invalid public key" pk))
  (unless (fx= (blob-size sk) asymmetric-box-secretkeybytes)
    (error 'exchange-random-stream-key "invalid secret key" sk))
  (let ([k (make-blob random-stream-keybytes)])
    (crypto-key-exchange k sk pk)
    (set-finalizer! k wipe!)
    (make-key random-stream-algorithm #f k)))

(define (exchange-symmetric-box-key pk sk)
  (set! pk
    (public-key (check-key 'exchange-symmetric-box-key asymmetric-box-keyalgorithm pk)))
  (set! sk
    (secret-key (check-key 'exchange-symmetric-box-key asymmetric-box-keyalgorithm sk)))
  (unless (fx= (blob-size pk) asymmetric-box-publickeybytes)
    (error 'exchange-symmetric-box-key "invalid public key" pk))
  (unless (fx= (blob-size sk) asymmetric-box-secretkeybytes)
    (error 'exchange-symmetric-box-key "invalid secret key" sk))
  (let ([k (make-blob symmetric-box-keybytes)])
    (crypto-key-exchange k sk pk)
    (set-finalizer! k wipe!)
    (make-key symmetric-box-algorithm #f k)))

(define (asymmetric-box pk sk)
  (symmetric-box (exchange-symmetric-box-key pk sk)))

(define (asymmetric-unbox pk sk)
  (symmetric-unbox (exchange-symmetric-box-key pk sk)))

(define (asymmetric-key-box pk sk)
  (symmetric-key-box (exchange-symmetric-box-key pk sk)))

(define (asymmetric-key-unbox pk sk)
  (symmetric-key-unbox (exchange-symmetric-box-key pk sk)))

)
;; vim: set ft=scheme sts=2 sw=2 ts=4 et ai: ;;
